﻿/****************************************************************************
//
//                I N F O S I G H T    C O R P O R A T I O N
//
//                ExtendedProtocolSynchronousClientSocket.cs
//
//		    This module contains class ExtendedProtocolSynchronousClientSocket
//      used to establish communications with an InfoSight marking system
//      controller over Ethernet / TCP/IP sockets. A public function is provided
//      to send Extended Protocol messages to the controller.
//
//      Note that this example uses Synchronous Sockets, meaning that program
//      execution will block while communications are in progress.
//
//
//   CREATED BY :      Joseph F. Muscarella
//                     InfoSight Corporation
//                     10/25/2016
//
//
//
//   MODIFICATION HISTORY
//   VERSION    DATE    INITIALS       DESCRIPTION
//   ======= ========== ========  ===========================================
//
//
//  This software is licensed under The MIT License (MIT)
//
//	Copyright (c) 2016  InfoSight Corporation
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy 
//  of this software and associated documentation files (the "Software"), to deal 
//  in the Software without restriction, including without limitation the rights 
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
//  of the Software, and to permit persons to whom the Software is furnished to do 
//  so, subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in all 
//  copies or substantial portions of the Software.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
//  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
//  PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
//  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
//  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
//  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
****************************************************************************/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;

namespace SampleExtendedProtocolApplication
{
    class ExtendedProtocolSynchronousClientSocket
    {
        /****************************************************************************
        //
        //   SendMessage
        //
        //   Syntax			:	void    SendMessage(remote_ip_address, remote_port, msg_type, msg_text, reply_text);
        //
        //   Arguments		:	string          remote_ip_address   - IP address of the marker (e.g. 192.168.1.3)
        //                      UInt16          remote_port         - The port to which the marker is listening (e.g. 5001)
        //                      byte            msg_type            - Extended Protocol message type ASCII character
        //                      string          msg_text            - The text string to send to the marker
        //                      out string      reply_text          - where we put any text reply from the marker.
        //
        //   Returns		:	nothing
        //
        //   Description	:	Send the message to the remote marking system using the
        //          InfoSight Extended Protocol over a synchronous (blocking) Socket.
        //
        ****************************************************************************/
        public  static  string  SendMessage(string remote_ip_address, UInt16 remote_port, byte msg_type, string msg_text, out string reply_text)
        {
            // Default status is OK.

            string  status_str  = "COM OK";

            // Clear the reply text as not all messages have text in the reply.

            reply_text  = "";

            try
            {
                // Create the socket that will connect to the remote marker's IP address / Port Number.

                IPAddress   ip_address      = IPAddress.Parse(remote_ip_address);
                IPEndPoint  remote_ep       = new IPEndPoint(ip_address, remote_port);
                Socket      client_socket   = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);

                try
                {
                    // Convert the input message text into an array of bytes.

                    byte[]      data_text   = Encoding.ASCII.GetBytes(msg_text);

                    // Use a byte List to build the Extended Protocol message format.

                    List<byte>  buffer      = new List<byte>();

                    buffer.Add((byte)AsciiCtrlTable.SOH);
                    buffer.Add(msg_type);
                    buffer.Add((byte)AsciiCtrlTable.STX);

                    // Add each byte of the message.  

                    for (int index = 0; index < data_text.Length; ++index)
                    {
                        buffer.Add(data_text[index]);
                    }

                    // Compute the checksum starting with the type and adding all the
                    // characters between the STX and ETX non-inclusive.

                    byte    sum = buffer[1];

                    for (int index = 3; index < buffer.Count(); ++index)
                    {
                        sum += buffer[index];
                    }

                    // Add the trailing end of text character.

                    buffer.Add((byte)AsciiCtrlTable.ETX);

                    // Now add the BCC as three decimal ASCII characters.
            
                    buffer.Add((byte)((sum / 100) + 0x30));
                    sum %= 100;
                    buffer.Add((byte)((sum / 10) + 0x30));
                    sum %= 10;
                    buffer.Add((byte)((sum) + 0x30));
            
                    // Add the final carriage return terminator character.

                    buffer.Add((byte)AsciiCtrlTable.CR);

                    // Connect the socket to the remote marker and send the buffer.

                    client_socket.Connect(remote_ep);
                    client_socket.Send(buffer.ToArray());

                    // Set a three second timeout on receiving the reply.

                    client_socket.ReceiveTimeout    = 3000;

                    // Get the reply into another byte buffer.

                    byte[]  recv_buffer = new byte[1024];
                    int     recv_count  = client_socket.Receive(recv_buffer);

                    // Decode the reply.
                    // Minimum of 6 bytes in a valid reply (may be more depending on the message)
                    // SOH  TYPE  ACK/NAK  STX  [data]  ETX  [BCC]  CR
                    // Note that the data field and BCC are optional.

                    if (recv_count >= 6)
                    {
                        // If the basic format is correct, extract the reply message and put it into the caller's variable.
                        // Note that this simple example doesn't decode the optional reply BCC characters.

                        if ((recv_buffer[0] == (byte)AsciiCtrlTable.SOH) &&
                            (recv_buffer[1] == msg_type) &&
                            (recv_buffer[2] == (byte)AsciiCtrlTable.ACK) &&
                            (recv_buffer[3] == (byte)AsciiCtrlTable.STX))
                        {
                            // Starting just after the STX character, extract characters until we encounter the ETX.

                            for (int index = 4; (index < recv_count) && (recv_buffer[index] != (byte)AsciiCtrlTable.ETX); ++index)
                            {
                                reply_text  = reply_text + (char)recv_buffer[index];
                            }
                        }
                        else if (recv_buffer[2] == (byte)AsciiCtrlTable.NAK)
                        {
                            status_str  = "NAK received.";
                        }
                        else if (recv_buffer[1] != msg_type)
                        {
                            status_str  = "Reply message type mismatch.";
                        }
                        else
                        {
                            status_str  = "Invalid reply message format.";
                        }
                    }
                    else
                    {
                        status_str  = "Invalid number of characters in the reply message.";
                    }

                    // With communications complete, shutdown and close the socket connection.

                    client_socket.Shutdown(SocketShutdown.Both);
                    client_socket.Close();
                 }
                catch ( ArgumentNullException ane )
                {
                    status_str  = "Argument Null Exception: " + ane.ToString();
                }
                catch ( SocketException se )
                {
                    status_str  = "Socket Exception: " + se.ToString();
                }
                catch ( Exception ue )
                {
                    status_str  = "Unexpected Exception: " + ue.ToString();
                }
            }
            catch ( Exception err )
            {
                status_str  = err.ToString();
            }

            return (status_str);
        }
    }

    // Define a table of ASCII control codes by name.

    public enum AsciiCtrlTable
    {
        NUL = 0,
        SOH = 1,
        STX = 2,
        ETX = 3,
        EOT = 4,
        ENQ = 5,
        ACK = 6,
        BEL = 7,
        BS  = 8,
        HT  = 9,
        LF  = 10,
        VT  = 11,
        FF  = 12,
        CR  = 13,
        SO  = 14,
        SI  = 15,
        DLE = 16,
        DC1 = 17,
        XON = 17,
        DC2 = 18,
        DC3 = 19,
        XOFF= 19,
        DC4 = 20,
        NAK = 21,
        SYN = 22,
        ETB = 23,
        CAN = 24,
        EM  = 25,
        SUB = 26,
        ESC = 27,
        FS  = 28,
        GS  = 29,
        RS  = 30,
        US  = 31,
        SP  = 32
    }

}
